/*-
 * Copyright (C)2018 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2018 @BABOLO http://www.babolo.ru/"
#ident "@(#) $Id: csv2pg.c,v 1.7 2019/09/04 19:03:18 babolo Exp $"

#define MIFE_COMPAT      5
#define BLIN_COMPAT      3
#define Bpars_COMPAT     4

#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>
#include <sysexits.h>
#include <stdlib.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include <mife.h>
#include "csv2pg.h"

#define Ce 0x00000100 /* write escaped  */
#define Co 0x00000200 /* write ordinar  */
#define Cq 0x00000400 /* write quote    */
#define Cf 0x00000800 /* fields control */
#define Ct 0x00002000 /* tuple ready    */
#define Cl 0x00004000 /* eol test       */
#define Cx 0x00800000 /* error          */
#define C_state  0xFF /* next state     */

static const char statenames[][3] = { "bg", "po", "pq", "qq", "el", "px" };
static const char clasnames[][3] = { "ce", "cc", "cq", "cl", "co", "xx" };
static const char subs[16] = {'0', 0, 0, 0 , 0, 0, 0, 0 , 'b', 't', 'n', 'v' , 'f', 'r', 0, 0};

enum states
{ bg  /*       */
, po  /*      */
, pq  /*       */
, qq  /*       */
, el  /*         */
, px
};

enum classes
{ ce  /* quoted  */
, cc  /* comma   */
, cq  /* quot    */
, cl  /* eol     */
, co  /* ordinar */
, xx
};

static const char pn[][3] = 
{ "- ", "- ", "- ", "- ", "- ", "- ", "- ", "- "
, "x ", "- ", "- ", "- ", "- ", "- ", "- ", "- "
, "- ", "l ", "t ", "n ", "f ", "q ", "o ", "e "
};

static enum classes class[256] =
{ ce, co, co, co , co, co, co, co , ce, ce, cl, cl , cl, cl, co, co
, co, co, co, co , co, co, co, co , co, co, co, co , co, co, co, co
, co, co, co, co , co, co, co, co , co, co, co, co , co, co, co, co
, co, co, co, co , co, co, co, co , co, co, co, co , co, co, co, co

, co, co, co, co , co, co, co, co , co, co, co, co , co, co, co, co
, co, co, co, co , co, co, co, co , co, co, co, co , co, co, co, co
, co, co, co, co , co, co, co, co , co, co, co, co , co, co, co, co
, co, co, co, co , co, co, co, co , co, co, co, co , co, co, co, co

, co, co, co, co , co, co, co, co , co, co, co, co , co, co, co, co
, co, co, co, co , co, co, co, co , co, co, co, co , co, co, co, co
, co, co, co, co , co, co, co, co , co, co, co, co , co, co, co, co
, co, co, co, co , co, co, co, co , co, co, co, co , co, co, co, co

, co, co, co, co , co, co, co, co , co, co, co, co , co, co, co, co
, co, co, co, co , co, co, co, co , co, co, co, co , co, co, co, co
, co, co, co, co , co, co, co, co , co, co, co, co , co, co, co, co
, co, co, co, co , co, co, co, co , co, co, co, co , co, co, co, co
};

static const u_int32_t automa[2][px][xx] =
/*       ce     cc     cq        cl     co     xx */
{ { { Ce|po, Cf|bg,    pq, Ct|Cl|el, Co|po }/* bg     */
  , { Ce|po, Cf|bg, Cx|px, Ct|Cl|el, Co|po }/* po    */
  , { Ce|pq, Co|pq,    qq, Ce   |pq, Co|pq }/* pq     */
  , { Cx|px, Cf|bg, Co|pq, Ct|Cl|el, Cx|px }/* qq     */
  , { Ce|po, Cf|bg,    pq,    Cl|el, Co|po }/* el       */
  }
, { { Ce|po, Cf|bg,    pq, Ct|Cl|el, Co|po }/* bg     */
  , { Ce|po, Cf|bg, Co|po, Ct|Cl|el, Co|po }/* po    */
  , { Ce|pq, Co|pq,    qq, Ce   |pq, Co|pq }/* pq     */
  , { Cq|pq, Cf|bg, Co|pq, Ct|Cl|el, Cq|pq }/* qq     */
  , { Ce|po, Cf|bg,    pq,    Cl|el, Co|po }/* el       */
} };

static int
csv2pg_do(csv2pg_cf *cf) {
    blin_statectl    sctl = {{4, 1, 8, 24, ' ', 3, 1, 0}};        /*  blin_stateheader()         */
    int              fieldcount;
    u_int32_t        control;
    off_t            offset;
    enum states      state;
    u_int32_t        lnctr;
    ssize_t          reed;
    int              ex  = EX_OK;
    char             ccl = cl;
    char            *p   = NULL;

#   define blin_internal_flags (cf->flags & BLIN_MASK)
    class[cf->flags & CVS2PG_QUOTE] = cq;
    class[(cf->flags & CVS2PG_COMMA) >> 8] = cc;
    if  (cf->flags & CVS2PG_XHEAD) cf->throw |= CVS2PG_THROW;
    blin_stateheader(sctl, pn);
    for ( cf->tuplecount = 0
        , fieldcount = 0
        , control = 0
        , offset = 0
        , state = bg
        , lnctr = 0
        , reed = 1
        ; (state < px) && !!reed
        ; ++offset
        ) {
        if  (0 > (reed = mife_read(cf->mdi, 1, offset))) {
            ifBLIN_QW0("mife_read");
            ex = EX_IOERR;
            goto out;
        }
        if  (!reed) {
            control = automa[!!(cf->flags & CVS2PG_RELAX)][state][cl];
            p = &ccl;
        } else if (!(p = mife_pointer(cf->mdi))) {
            ifBLIN_QW0("mife_pointer");
            ex = EX_IOERR;
            goto out;
        } else {
            control = automa[!!(cf->flags & CVS2PG_RELAX)][state][class[*p & 0xFF]];
        }
        blin_statebody( sctl
                      , pn
                      , statenames
                      , clasnames[class[*p & 0xFF]]
                      , p
                      , offset
                      , control
                      , state
                      , fieldcount
                      , p[0]
                      );
        if  (!cf->throw) {
            if  (control & Ce) {
                if  (0 > mife_aput(cf->out, '\\')) {
                    ifBLIN_QW0("mife_aput \\");
                    ex = EX_IOERR;
                    goto out;
                }
                if  (0 > mife_aput(cf->out, subs[*p & 0xFF])) {
                    ifBLIN_QW0("mife_aput");
                    ex = EX_IOERR;
                    goto out;
            }   }
            if  (control & Co) {
                if  (0 > mife_aput(cf->out, *p)) {
                    ifBLIN_QW0("mife_aput");
                    ex = EX_IOERR;
                    goto out;
            }   }
            if  (control & Cq) {
                if  (0 > mife_aput(cf->out, cf->flags & CVS2PG_QUOTE)) {
                    ifBLIN_QW0("mife_aput %c");
                    ex = EX_IOERR;
                    goto out;
        }   }   }
        if  (control & Cf) {
            ++fieldcount;
            if  ((cf->fieldnom >= 0) && (fieldcount >= cf->fieldnom)) {
                if  (cf->flags & (CVS2PG_OVEVA | CVS2PG_UNEVA)) {
                    ifBLIN_QX1( "String %d: field %d ,  %d  XXXX"
                              , cf->tuplecount
                              , fieldcount
                              , cf->fieldnom
                              );
                    if  (cf->flags & CVS2PG_OVEVA) cf->throw |= CVS2PG_THCUT;
                } else {
                    ifBLIN_QX0( "String %d: field %d ,  %d  XXXX"
                              , cf->tuplecount
                              , fieldcount
                              , cf->fieldnom
                              );
                    ex = EX_DATAERR;
                    errno = ERANGE;
                    goto out;
            }   }
            if  ((!cf->throw) && (0 > mife_aput(cf->out, '	'))) {
                ifBLIN_QW0("mife_aput");
                ex = EX_IOERR;
                goto out;
        }   }
        if  (  (control & Ct)
            && (  !(cf->flags & CVS2PG_SCEVA)
               || (cf->fieldnom < 0)
               || ((fieldcount + 1) >= cf->fieldnom)
            )  ) {
            if  (cf->fieldnom < 0) {
                cf->fieldnom = fieldcount + 1;
            } else if (cf->flags & CVS2PG_OVEVA) {
                while (!cf->throw && (fieldcount < cf->fieldnom)) {
                    if  (++fieldcount >= cf->fieldnom) break;
                    if  (0 > mife_aput(cf->out, '	')) {
                        ifBLIN_QW0("mife_aput");
                        ex = EX_IOERR;
                        goto out;
                    }
                    if  (0 > mife_aput(cf->out, '\\')) {
                        ifBLIN_QW0("mife_aput \\");
                        ex = EX_IOERR;
                        goto out;
                    }
                    if  (0 > mife_aput(cf->out, 'N')) {
                        ifBLIN_QW0("mife_aput");
                        ex = EX_IOERR;
                        goto out;
                }   }
            } else if (++fieldcount < cf->fieldnom) {
                if  (!(cf->flags & CVS2PG_UNEVA)) {
                    ifBLIN_QX0( "short tuple %d: last field %d of %d"
                              , cf->tuplecount
                              , fieldcount
                              , cf->fieldnom
                              );
                    ex = EX_DATAERR;
                    errno = ERANGE;
                    goto out;
                } else {
                    ifBLIN_QX1( "short tuple %d: last field %d of %d"
                              , cf->tuplecount
                              , fieldcount
                              , cf->fieldnom
                              );
            }   }
            cf->tuplecount++;
            fieldcount = 0;
            lnctr = 0;
            cf->throw &= ~CVS2PG_THCUT;
            if  (!cf->throw) {
                if  (0 > mife_aput(cf->out, '\n')) {
                    ifBLIN_QW0("mife_aput");
                    ex = EX_IOERR;
                    goto out;
            }   }
            cf->throw &= ~CVS2PG_THROW;
        }
        if  (control & Cl) {
            if  (cf->flags & CVS2PG_OHEAD) break;
            if  ((1 << (*p & 0xFF)) & lnctr) {
                ifBLIN_QX1("Empty string N %d", cf->tuplecount);
            }
            lnctr |= 1 << (*p & 0xFF);
        }
        if  (control & Cx) {
            ifBLIN_QX0("Syntax error at line %d field %d", cf->tuplecount, fieldcount);
            goto out;
        }
        state = control & C_state;
    }
    if  (!ex && (cf->flags & CVS2PG_COEOF)) {
        if  (0 > mife_aput(cf->out, '\\')) {
            ifBLIN_QW0("mife_aput");
            ex = EX_IOERR;
            goto out;
        }
        if  (0 > mife_aput(cf->out, '.')) {
            ifBLIN_QW0("mife_aput");
            ex = EX_IOERR;
            goto out;
        }
        if  (0 > mife_aput(cf->out, '\n')) {
            ifBLIN_QW0("mife_aput");
            ex = EX_IOERR;
            goto out;
    }   }
out:
    if  (0 > mife_aclose(cf->out)) {
        ifBLIN_QW0("mife_aclose");
        ex = EX_IOERR;
    }
    return(ex);
#   undef blin_internal_flags
}

static void
csv2pg_dumpopts(csv2pg_cf *cf) {
#   define blin_internal_flags (cf->flags & BLIN_MASK)
    printf("       quote	-w%c\n", cf->flags & 0x00FF);
    printf("       quote	-W%02X\n", cf->flags & 0x00FF);
    printf("       comma	-c%c\n", (cf->flags >> 8) & 0x00FF);
    printf("       comma	-C%02X\n", (cf->flags >> 8) & 0x00FF);
    if  (cf->flags & CVS2PG_UNEVA) printf("    unallign	-u\n");
    if  (cf->flags & CVS2PG_OVEVA) printf("      allign	-x\n");
    if  (cf->flags & CVS2PG_COEOF) printf("         eof	-e\n");
    if  (cf->flags & CVS2PG_RELAX) printf("  relaxquote	-r\n");
    if  (cf->flags & CVS2PG_WHENY) printf("     sizeout	-N\n");
    if  (cf->flags & CVS2PG_XHEAD) printf("     headcut	-h\n");
    if  (cf->flags & CVS2PG_OHEAD) printf("    headonly	-H\n");
    if  (cf->fieldnom >= 0) printf("   fieldsnum	-n%d\n", cf->fieldnom);
    if  (cf->inname) printf("       input	-f%s\n", cf->inname);
    return;
#   undef blin_internal_flags
}

static void
usage(int ex) {
    fprintf( stderr
           , "csv2pg @BABOLO V.M "VERS"  "DATE" http://www.babolo.ru/\n"
             "    -c SYM    - comma symbol literal (default ,)\n"
             "    -C CODE   - comma symbol code (default 2C)\n"
             "    -e        - \\. at end\n"
             "    -g        - glue short lines together\n"
             "    -h        - header delete\n"
             "    -H        - header only\n"
             "    -f FILE   - from file, default stdin\n"
             "    -n NUMBER - number of fields\n"
             "    -N        - print number of fields and tuples\n"
             "    -o FILE   - to file, default stdout\n"
             "    -q        - quiet\n"
             "    -r        - allow single quote inside field\n"
             "    -u        - get fields number as is\n"
             "    -v        - verbose\n"
             "    -w SYM    - quote symbol literal (default \")\n"
             "    -W CODE   - quote symbol code (default 5C)\n"
             "    -x        - align field number\n"
             "    -?        - this text\n"
           );
    exit(ex);
}

int
main(int argc, char *argv[]) {
    const char      *flar = "c:C:eghHf:n:No:qruvw:W:x=:?";
    babolo_opts     *bos = NULL;
    int              tfd;
    int              ex = EX_OK;
    csv2pg_cf       *cf;
    const char      *o;
    int              c;

    if  (!(cf = calloc(1, sizeof(csv2pg_cf)))) {
        ifBLIN_QW0("No mem %"BLIN_X, sizeof(csv2pg_cf));
        errno = ENOMEM;
        ex = EX_OSERR;
        goto out;
    }
    cf->flags = BLIN_BIT0;
#   define blin_internal_flags (cf->flags & BLIN_MASK)
    cf->fieldnom = -1;
    cf->flags |= '\"' & 0xFF;
    cf->flags |= (',' & 0xFF) << 8;
    if  (!(bos = babolo_openopts(BLIN_BIT0 | Bpars_NONU | Bpars_SYME | Bpars_ESYM, 'Z'))) {
        ifBLIN_QW0("babolo_openopts");
        ex = EX_SOFTWARE;
        goto out;
    }
    if  (babolo_setopts(bos, Bpars_NONU, argc, argv, flar)) {
        ifBLIN_QW0("babolo_setopts");
        ex = EX_SOFTWARE;
        goto out;
    }
    while (0 < (c = babolo_getopts(bos))) {
        switch (c) {
        case 'c': if  (((cf->flags >> 8) & 0xFF) != (',' & 0xFF))  ifBLIN_QX1("Comma redefined");
                ; if  (!(o = babolo_getoptsarg(bos))) {
                ;     ifBLIN_QW0("No -%c arg", bos->c);
                ;     ex = EX_USAGE;
                ;     goto out;
                ; }

                ; cf->flags = (cf->flags & ~0x00FF00) | ((*o << 8) & 0x00FF00);
                ; break;
        case 'C': if  (((cf->flags >> 8) & 0xFF) != (',' & 0xFF))  ifBLIN_QX1("Comma redefined");
                ; errno = 0;
                ; cf->flags = (cf->flags & ~0x00FF00) | ((babolo_getoptunum(bos, 0) << 8) & 0x00FF00);
                ; if  (errno) {
                ;     ifBLIN_QW0("-%c babolo_getoptunum", bos->c);
                ;     ex = EX_USAGE;
                ;     goto out;
                ; }
                ; break;
        case 'e': cf->flags |= CVS2PG_COEOF;
                ; break;
        case 'g': cf->flags |= CVS2PG_SCEVA;
                ; break;
        case 'h': cf->flags |= CVS2PG_XHEAD;
                ; break;
        case 'H': cf->flags |= CVS2PG_OHEAD;
                ; break;
        case 'f': if  (!!cf->mdi) ifBLIN_QX1("Input file redefined");
                ; if  (!(o = babolo_getoptsarg(bos))) {
                ;     ifBLIN_QW0("No -%c arg", bos->c);
                ;     ex = EX_USAGE;
                ;     goto out;
                ; }
                ; if  (!(cf->mdi = mife_init(BLIN_BIT0 | MIFE_SLID))) {
                ;     ifBLIN_QW0("mife_init");
                ;     ex = EX_SOFTWARE;
                ;     goto out;
                ; }
                ; if  (0 > mife_ctlfile(cf->mdi, o)) {
                ;     ifBLIN_QW0("mife_ctlfile %s", o);
                ;     ex = EX_IOERR;
                ;     goto out;
                ; }
                ; cf->inname = o;
                ; break;
        case 'n': errno = 0;
                ; if  (0 < cf->fieldnom) ifBLIN_QX1("Field number redefined");
                ; cf->fieldnom = babolo_getoptunum(bos, 0);
                ; if  (errno) {
                ;     ifBLIN_QW0("-%c babolo_getoptunum", bos->c);
                ;     ex = EX_USAGE;
                ;     goto out;
                ; }
                ; break;
        case 'N': cf->flags |= CVS2PG_WHENY;
                ; break;
        case 'o': if  (!!cf->out) ifBLIN_QX1("Output file redefined");
                ; if  (!(o = babolo_getoptsarg(bos))) {
                ;     ifBLIN_QW0("No -%c arg", bos->c);
                ;     ex = EX_USAGE;
                ;     goto out;
                ; }
                ; if  (0 > (tfd = open(o, O_WRONLY | O_CREAT | O_TRUNC, 0666))) {
                ;     ifBLIN_QW0("open %s", o);
                ;     ex = EX_IOERR;
                ;     goto out;
                ; }
                ; if  (!(cf->out = mife_aopen(BLIN_BIT0, tfd))) {
                ;     ifBLIN_QW0("mife_aopen %s", o);
                ;     ex = EX_IOERR;
                ;     goto out;
                ; }
                ; break;
        case 'q': cf->flags &= ~BLIN_MASK;
                ; break;
        case 'r': cf->flags |= CVS2PG_RELAX;
                ; break;
        case 'u': cf->flags |= CVS2PG_UNEVA;
                ; break;
        case 'v': BLIN_VERBOSE(cf->flags);
                ; break;
        case 'w': if  ((cf->flags & 0xFF) != ('\"' & 0xFF))  ifBLIN_QX1("Quote redefined");
                ; if  (!(o = babolo_getoptsarg(bos))) {
                ;     ifBLIN_QW0("No -%c arg", bos->c);
                ;     ex = EX_USAGE;
                ;     goto out;
                ; }
                ; cf->flags = (cf->flags & ~0x00FF) | (*o & 0x00FF);
                ; break;
        case 'W': if  ((cf->flags & 0xFF) != ('\"' & 0xFF))  ifBLIN_QX1("Quote redefined");
                ; errno = 0;
                ; cf->flags = (cf->flags & ~0x00FF) | (babolo_getoptunum(bos, 0) & 0x00FF);
                ; if  (errno) {
                ;     ifBLIN_QW0("-%c babolo_getoptunum", bos->c);
                ;     ex = EX_USAGE;
                ;     goto out;
                ; }
                ; break;
        case 'x': cf->flags |= CVS2PG_OVEVA;
                ; break;
        case '=': if  (!(o = babolo_getoptsarg(bos))) {
                ;     ifBLIN_QW0("No -%c arg", bos->c);
                ;     ex = EX_USAGE;
                ;     goto out;
                ; }
                ; switch(o[0]) {
                  case '0': blin_ctl(BLIN_CTL_DUMP);           exit(EX_OK);
                  case '1': babolo_optest(flar);               exit(EX_OK);
                  case '2': babolo_optext(bos);                exit(EX_OK);
                  case '3': babolo_dumpopts(bos);              exit(EX_OK);
                  case '=': csv2pg_dumpopts(cf);               exit(EX_OK);
                  }
                ; /* FALLTHROUGH */
        case '?': usage(EX_OK);
        default : ifBLIN_QW0("Ill flag -%c", bos->c);
                ; usage(EX_USAGE);
    }   }
    if  (0 > c) usage(EX_USAGE);
    if  (!cf->mdi) {
        if  (!(cf->mdi = mife_init(MIFE_SLID))) {
            ifBLIN_QW0("mife_init");
            ex = EX_IOERR;
            goto out;
        }
        if  (0 > mife_ctlfdsc(cf->mdi, fileno(stdin))) {
            ifBLIN_QW0("mife_ctlfdsc stdin");
            ex = EX_IOERR;
            goto out;
        }
        cf->inname = "/dev/stdin";
    }
    if  (!cf->out) {
        if  (!(cf->out = mife_aopen(0, fileno(stdout)))) {
            ifBLIN_QW0("mife_aopen stdout");
            ex = EX_IOERR;
            goto out;
    }   }
    if  (0 > (ex = csv2pg_do(cf))) ifBLIN_QW0("csv2pg_do");
    if  (!ex && (cf->flags & CVS2PG_WHENY)) {
        fprintf( stderr, "%s\t%d\tfields\n%s\t%d\ttuples\n"
               , cf->inname, cf->fieldnom
               , cf->inname, cf->tuplecount
               );
    }
out:
    exit(ex);
#   undef blin_internal_flags
}
